import datetime

import google.auth
import google.auth.compute_engine
import google.auth.transport.requests
from django.conf import settings
from django.core.cache import cache
from django.utils.deconstruct import deconstructible
from storages.backends.gcloud import GoogleCloudStorage


@deconstructible
class BaseStorage(GoogleCloudStorage):
    CACHE_KEY = "GoogleCloudStorageAccessToken.signing_extra_params"

    def __init__(self, dev_override=True, *args, **kwargs):
        # Override the bucket for non production environments
        if dev_override and not settings.PROD:
            kwargs["bucket_name"] = "dev-bucket.taficasa.com"
        super().__init__(*args, **kwargs)

    def url(self, name, parameters=None):
        """
        This function loads in extra credential params to
        the call for signing urls

        It is a workaround for not being able to
        use service accounts to sign urls.

        Workaround inferred from github issue below
        -> https://github.com/jschneier/django-storages/issues/941#issuecomment-797724786
        """

        # We only need this workaround when running in cloud run where we have a
        # service account attached.
        if not settings.LOCAL or settings.TESTING:
            extra_params = self.get_extra_signing_credentials()
            return super().url(name, parameters=extra_params)
        else:
            return super().url(name)

    def get_extra_signing_credentials(self):
        value = cache.get(self.CACHE_KEY)
        if value is not None:
            expiry, extra_params = value
            if expiry > datetime.datetime.utcnow():
                return extra_params

        credentials, project_id = google.auth.default()
        auth_req = google.auth.transport.requests.Request()
        credentials.refresh(auth_req)
        extra_params = {
            "service_account_email": credentials.service_account_email,
            "access_token": credentials.token,
            "credentials": credentials,
        }

        cache.set(self.CACHE_KEY, (credentials.expiry, extra_params))
        return extra_params


@deconstructible
class MediaStorage(BaseStorage):
    def __init__(self, *args, **kwargs):
        kwargs["bucket_name"] = "media.taficasa.com"
        super().__init__(*args, **kwargs)


@deconstructible
class StaticFilesStorage(BaseStorage):
    def __init__(self, *args, **kwargs):
        if settings.PROD:
            kwargs["bucket_name"] = "static.taficasa.com"
        else:
            kwargs["bucket_name"] = "dev.static.taficasa.com"
        super().__init__(dev_override=False, *args, **kwargs)


@deconstructible
class SupplyChainStorage(BaseStorage):
    def __init__(self, *args, **kwargs):
        kwargs["bucket_name"] = "supplychain.taficasa.com"
        super().__init__(*args, **kwargs)


@deconstructible
class OrganizationsStorage(BaseStorage):
    def __init__(self, *args, **kwargs):
        kwargs["bucket_name"] = "organizations.taficasa.com"
        super().__init__(*args, **kwargs)



import random
import string
from typing import List, Set, Tuple

import names
from authentication.users.models import User
from core.testing.constants import DEFAULT_TEST_PASSWORD
from django.contrib.auth import get_user_model
from organization_management.organizations.models import Organization
from organization_management.workflows.constants import (
    WorkflowSchemaUseCaseType,
    WorkflowStepType,
    WorkflowTargetObject,
)
from organization_management.workflows.models import WorkflowSchemaVersion
from organization_management.workflows.services import (
    WorkflowSchemaService,
    WorkflowService,
)
from pyre_extensions import none_throws
from supply_chain.merchant_registrations.constants import (
    MerchantRegistrationConfigurationStepType,
)
from supply_chain.merchant_registrations.models import (
    MerchantRegistrationConfigurationVersion,
)
from supply_chain.merchant_registrations.services import (
    SupplyChainMerchantRegistrationService,
)
from supply_chain.products.models import Product
from supply_chain.profiles.constants import SupplyChainOrganizationType
from supply_chain.profiles.models import SupplyChainProfile


class SupplyChainDataMixin:
    created_users: Set
    created_products: Set

    def get_test_supply_chain_profile(
        self, profile_type: SupplyChainOrganizationType
    ) -> Tuple[User, SupplyChainProfile]:
        # Import services in functions to avoid circular imports
        from organization_management.organizations.services import OrganizationService
        from supply_chain.profiles.services import SupplyChainProfileService

        # Create a new user
        random_name = names.get_full_name()
        while random_name in self.created_users:
            random_name = names.get_full_name()

        firstname, lastname = random_name.split(" ")
        user_data = {
            "email": f"{firstname}{lastname}@test.com",
            "firstname": firstname,
            "lastname": lastname,
            "password": DEFAULT_TEST_PASSWORD,
        }

        # Create User to be used in tests
        user: User = get_user_model().objects.create_user(**user_data)
        user.email_validated = True
        user.save()

        organization = OrganizationService().create_organization(
            "Test Organization",
            user,
            location_data={
                "street_address": "No 1 lokogoma street",
                "country": "NG",
                "locality": "Wuse",
                "administrative_area": "Abuja",
            },
        )

        organization.is_verified = True
        organization.save()

        # Update the created users set
        self.created_users.add(random_name)

        # Set up a supply chain profiles needed for tests
        profile = none_throws(
            SupplyChainProfileService.create_profile(
                organization=organization,
                created_by=user,
                organization_type=profile_type,
                product_receipt_prefix="TEST"
                if profile_type == SupplyChainOrganizationType.MERCHANT
                else None,
            )
        )

        return user, profile

    def get_test_product(
        self, merchant: SupplyChainProfile, created_by: User
    ) -> Product:
        # Import services in functions to avoid circular imports
        from supply_chain.products.services import ProductService

        # Get a random product id
        product_id = "".join(random.choices(string.digits, k=5))
        while product_id in self.created_products:
            product_id = "".join(random.choices(string.digits, k=5))

        # First create workflow schema and workflow to create product
        self.create_test_workflow_schema(
            published_by=created_by,
            organization=merchant.organization,
            use_case=WorkflowSchemaUseCaseType.PRODUCT_PUBLISH,
        )

        product = ProductService.create_product(
            merchant=merchant,
            name="Test Product",
            unit="bags",
            merchant_product_id=product_id,
            created_by=created_by,
        )

        # Complete approval workflow for product
        WorkflowService.auto_complete_workflow(
            workflow=product.publish_approval_workflow,
            target_object=WorkflowTargetObject.NONE,
            target_object_id="1",
            use_case=WorkflowSchemaUseCaseType.PRODUCT_PUBLISH,
        )

        product = ProductService.publish_product(
            product=product, published_by=created_by
        )
        return product

    def create_test_workflow_schema(
        self,
        published_by: User,
        organization: Organization,
        use_case: WorkflowSchemaUseCaseType,
    ) -> WorkflowSchemaVersion:
        steps_data = {
            "steps": [
                {
                    "step_title": "Owner Review Step",
                    "step_order": 1,
                    "step_description": "Generated step for owner review.",
                    "step_type": WorkflowStepType.REVIEW,
                },
                {
                    "step_title": "Owner Approval Step",
                    "step_order": 2,
                    "step_description": "Generated step for owner approval.",
                    "step_type": WorkflowStepType.APPROVAL,
                },
            ],
        }

        workflow_schema_version = WorkflowSchemaService.create_or_update_schema_version(
            organization=organization,
            version_id=1,
            steps=steps_data["steps"],
            use_case=use_case,
            published_by=published_by,
            should_publish=True,
            updated_by=published_by,
        )

        return workflow_schema_version

    def create_test_merchant_registration_configuration(
        self,
        profile: SupplyChainProfile,
        created_by: User,
    ) -> MerchantRegistrationConfigurationVersion:
        steps = [
            {
                "name": "Generated name",
                "step_type": MerchantRegistrationConfigurationStepType.OTHER,
                "step_order": 1,
                "description": "Generated description",
            }
        ]

        configuration = SupplyChainMerchantRegistrationService.create_configuration(
            profile, steps, created_by
        )

        return SupplyChainMerchantRegistrationService.get_current_version(configuration)



import uuid

from bson import ObjectId
from bson.errors import InvalidId
from django.apps import apps
from django.contrib.auth import get_user_model
from django.utils.encoding import smart_str
from rest_framework import serializers


class UserIdValidator:
    """
    This validator checks if the userid passed in is valid
    and if so returns a user object. This validator requires
    that context be passed into the serializer with the request
    instance.

    Should be applied to an individual field
    """

    requires_context = True

    def __init__(self, check_organization: bool = False):
        self.check_organization = check_organization

    def __call__(self, value, serializer_field):
        pass


class UserIdSerializerField(serializers.UUIDField):
    def __init__(self, check_organization: bool = False, **kwargs):
        self.check_organization = check_organization
        super().__init__(**kwargs)

    def to_representation(self, value):
        if hasattr(value, "userid"):
            return str(value.userid)
        else:
            return str(value)

    def to_internal_value(self, data):
        # Lazy loading all models needed
        Organization = apps.get_model(
            app_label="organization_management", model_name="Organization"
        )

        try:
            if isinstance(data, str):
                uuid_value = uuid.UUID(hex=data)
            else:
                raise serializers.ValidationError(
                    f"This userid '{data}' is not a uuid string"
                )

            user = get_user_model().objects.filter(userid=uuid_value).get()

            if self.check_organization:
                request = self.context.get("request")
                if not request:
                    raise serializers.ValidationError(
                        "Request must be passed into the serializer context"
                        " to use the UserIdSerializerField field"
                    )

                # Check the request url captured kwargs and request data
                organiaztion_id = request.data.get(
                    "organization_id"
                ) or request.resolver_match.kwargs.get("organization_id")

                if not organiaztion_id:
                    raise serializers.ValidationError("Organization ID not provided")

                try:
                    organization = Organization.objects.filter(
                        organization_id=organiaztion_id
                    ).get()
                    if not organization.is_member(user):
                        raise serializers.ValidationError(
                            f"This userid '{data}' does not belong to the"
                            " organization being accessed"
                        )

                except Organization.DoesNotExist:
                    raise serializers.ValidationError(
                        "Organization being accessed is invalid"
                    )
            return user
        except get_user_model().DoesNotExist:
            raise serializers.ValidationError(f"This userid '{data}' is invalid")


class ObjectIdField(serializers.Field):
    """Field for ObjectId values"""

    def to_internal_value(self, value):
        try:
            return ObjectId(smart_str(value))
        except InvalidId:
            raise serializers.ValidationError("'%s' is not a valid ObjectId" % value)

    def to_representation(self, value):
        return smart_str(value)



# Inspired by https://gist.github.com/hakib/e2e50d41d19a6984dc63bd94580c8647

"""
Custom django checks.
    H001: Field has no verbose name.
    H002: Model must define class Meta.
    H003: Model must define db table name.
    H004: Model must define db table name comment.
"""
import ast
import inspect

import django.apps
from django.conf import settings
from django.core import checks
from django.core.exceptions import FieldDoesNotExist


def get_argument(node, arg):
    for kw in node.value.keywords:
        if kw.arg == arg:
            return kw
    return None


def check_model(model):
    """Check a single model.
    Yields (django.checks.CheckMessage)
    """
    model_source = inspect.getsource(model)
    model_node = ast.parse(model_source)
    assert isinstance(model_node, ast.Module)

    class_meta = None
    for node in model_node.body[0].body:  # type: ignore
        if isinstance(node, ast.ClassDef):
            # class Meta
            if node.name == "Meta":
                class_meta = node

        # Fields
        elif isinstance(node, ast.Assign):
            if len(node.targets) != 1:
                continue

            if not isinstance(node.targets[0], ast.Name):
                continue

            field_name = node.targets[0].id
            try:
                field = model._meta.get_field(field_name)
            except FieldDoesNotExist:
                continue

            verbose_name = get_argument(node, "verbose_name")
            if verbose_name is None:
                yield django.core.checks.Warning(
                    "Field has no verbose name",
                    hint="Set verbose name on the field.",
                    obj=field,
                    id="H001",
                )

    if class_meta is None:
        yield django.core.checks.Warning(
            'Model "{}" must define class Meta'.format(model._meta.model_name),
            hint='Add class Meta to model "{}".'.format(model._meta.model_name),
            obj=model,
            id="H002",
        )

    else:
        db_table = None
        db_table_comment = None

        for node in ast.iter_child_nodes(class_meta):
            if not isinstance(node, ast.Assign):
                continue

            if not isinstance(node.targets[0], ast.Name):
                continue

            attr = node.targets[0].id

            if attr == "db_table":
                db_table = node

            if attr == "db_table_comment":
                db_table_comment = node

        if db_table is None:
            yield django.core.checks.Warning(
                "Model has no database table name",
                hint="Add db_table to class Meta.",
                obj=model,
                id="H003",
            )

        if db_table_comment is None:
            yield django.core.checks.Warning(
                "Model has no database table name comment",
                hint="Add db_table_comment to class Meta.",
                obj=model,
                id="H004",
            )


@checks.register(django.core.checks.Tags.models)
def check_models(app_configs, **kwargs):
    errors = []
    for app in django.apps.apps.get_app_configs():
        # Skip third party apps.
        if app.path.find("site-packages") > -1:
            continue

        for model in app.get_models():
            # Skip all models that start with the string 'Historical'
            # since they are autogenerated by django simple history
            if model.__name__.startswith("Historical"):
                continue

            for check_message in check_model(model):
                errors.append(check_message)

    return errors


@checks.register()
def check_migrate_run(app_configs, **kwargs):
    errors = []
    # Only run this check for specific django db related commands
    if (
        settings.CURRENT_DJANGO_COMMAND in {"migrate", "makemigrations"}
        and settings.USE_CLOUD_SQL_AUTH_PROXY
    ):
        errors.append(
            checks.Error(
                "Cannot run database commands e.g. migrate or makemigrations against any cloud databases",
                hint=(
                    "Please make sure USE_CLOUD_SQL_AUTH_PROXY is set to False before"
                    "running any database commands"
                ),
            )
        )

    return errors



from collections import OrderedDict

from after_effects.registry import after_effect
from after_effects.serializers import AfterEffectSerializer
from core.routers import CustomAPIRoute
from rest_framework import mixins, permissions
from rest_framework import status as http_status
from rest_framework import viewsets
from rest_framework.response import Response


class AfterEffectRunViewset(
    mixins.CreateModelMixin,
    viewsets.GenericViewSet,
):
    """
    A view to allow cloud tasks to run after effects jobs
    """

    serializer_class = AfterEffectSerializer
    permission_classes = [permissions.AllowAny]
    custom_api_routes = [
        CustomAPIRoute(
            url="run",
            url_args=OrderedDict(),
            url_name="after_effects_run",
            methods=["create"],
        )
    ]

    def create(self, request, *args, **kwarg):
        # Validate request data
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        # Run task
        effect = after_effect.registered_effects[serializer.validated_data["effect"]]
        effect(
            *serializer.validated_data["effect_args"],
            **serializer.validated_data["effect_kwargs"],
        )

        return Response(status=http_status.HTTP_200_OK)



from core.checks.models import check_migrate_run, check_models
from core.checks.permissions import run_access_control_permissions_check

__all__ = ["check_models", "check_migrate_run", "run_access_control_permissions_check"]



from typing import Any, Dict, List


def transform_list(data: List, id_field: str) -> Dict[str, Any]:
    """
    This function takes a list of objects and transforms them into
    objects where the key is the id (noted by the id_field) and the
    value is the object
    """

    transformed_data = {}
    for entity in data:
        entity_id = entity[id_field]
        transformed_data[entity_id] = entity
    return transformed_data



import os
from typing import Dict

import sendgrid  # type: ignore
from core.email.schema import SendgridVerifiedSender, SengridEmailTemplate
from django.conf import settings
from sendgrid.helpers.mail import Mail, To


class EmailClientSendGrid:
    """
    A class used to send predefined emails using the
    SendGrid api.
    """

    def __init__(self):
        api_key = os.getenv("SENDGRID_API_KEY", None)
        self.client = sendgrid.SendGridAPIClient(api_key=api_key) if api_key else None
        self.max_retries = 3

    def _create_mail(self, mail_data: Dict[str, str]) -> Mail:
        # If this is local dev forward all emails to the
        # email dev address
        if settings.LOCAL:
            # Get current dev email
            if settings.LOCAL_GCLOUD_CONFIG:
                current_dev_email = settings.LOCAL_GCLOUD_CONFIG["active_account"]
            else:
                current_dev_email = "email-dev@taficasa.com"
            mail_data["to_emails"] = current_dev_email

        return Mail(
            from_email=mail_data["from_email"],
            to_emails=mail_data["to_emails"],
        )

    def _send(self, message: Mail):
        # Do not send mail during tests or if client is None
        if not self.client or settings.TESTING:
            return

        try_send_mail = True
        mail_send_tries = 0
        while try_send_mail:
            try:
                mail_send_tries += 1
                response = self.client.send(message)

                if response.status_code == 202 or mail_send_tries >= self.max_retries:
                    try_send_mail = False
            except Exception as e:
                try_send_mail = False
                # TODO Add logging functionality for any email failures
                pass

    def send_welcome_mail(self, user):
        message = self._create_mail(
            {
                "from_email": (SendgridVerifiedSender.SUPPORT),
                "to_emails": [To(email=user.email, name=user.full_name)],
            },
        )

        # Set transactional template id and data for the welcome mails
        message.template_id = SengridEmailTemplate.WELCOME_EMAIL
        message.dynamic_template_data = {"firstname": user.firstname}
        self._send(message)

    def send_welcome_mail_on_proxy_creation(
        self, user, creator, organization, temp_password
    ):
        message = self._create_mail(
            {
                "from_email": (SendgridVerifiedSender.SUPPORT),
                "to_emails": [To(email=user.email, name=user.full_name)],
            },
        )

        # Set transactional template id and data for the welcome mails
        message.template_id = SengridEmailTemplate.PROXY_WELCOME
        message.dynamic_template_data = {
            "firstname": user.firstname,
            "temp_password": temp_password,
            "creator_firstname": creator.firstname,
            "org_name": organization.name,
        }
        self._send(message)

    def send_password_reset_request_mail(self, user, reset_url):
        message = self._create_mail(
            {
                "from_email": (SendgridVerifiedSender.SUPPORT),
                "to_emails": [To(email=user.email, name=user.full_name)],
            },
        )

        # Set transactional template id and data for the password reset request mails
        message.template_id = SengridEmailTemplate.PASSWORD_RESET
        message.dynamic_template_data = {
            "firstname": user.firstname,
            "reset_url": reset_url,
        }
        self._send(message)

    def send_password_changed_mail(self, user):
        message = self._create_mail(
            {
                "from_email": (SendgridVerifiedSender.SUPPORT),
                "to_emails": [To(email=user.email, name=user.full_name)],
            },
        )

        # Set transactional template id and data for the password reset request mails
        message.template_id = SengridEmailTemplate.PASSWORD_CHANGED
        message.dynamic_template_data = {"firstname": user.firstname}

        self._send(message)

    def send_email_verification_mail(self, user, verify_url):
        message = self._create_mail(
            {
                "from_email": (SendgridVerifiedSender.SUPPORT),
                "to_emails": [To(email=user.email, name=user.full_name)],
            },
        )

        # Set transactional template id and data for the welcome mails
        message.template_id = SengridEmailTemplate.EMAIL_VERIFICATION
        message.dynamic_template_data = {
            "firstname": user.firstname,
            "verify_url": verify_url,
        }

        self._send(message)

    def send_verification_acknowledgement_mail(self, user):
        """Method for sending email when an organization verification request has been made."""
        message = self._create_mail(
            {
                "from_email": (SendgridVerifiedSender.SUPPORT),
                "to_emails": [To(email=user.email, name=user.full_name)],
            },
        )

        # Use this while we design a template for pending verifications
        message.template_id = SengridEmailTemplate.WELCOME_EMAIL
        message.dynamic_template_data = {"firstname": user.firstname}
        self._send(message)

    def send_verification_approval_mail(self, user):
        """Method for sending email to a user when an organization has been verified."""
        message = self._create_mail(
            {
                "from_email": (SendgridVerifiedSender.SUPPORT),
                "to_emails": [To(email=user.email, name=user.full_name)],
            },
        )

        # Use this while we design a template for approved verifications
        message.template_id = SengridEmailTemplate.WELCOME_EMAIL
        message.dynamic_template_data = {"firstname": user.firstname}
        self._send(message)



from unittest import TestCase, mock, skip

from core.data.gen import generate_basic_data


def mock_save(*args, **kwargs):
    return "Fake file"


@skip("Skipping until we add logs that do not show in test for the populate functions")
class PopulateDataTests(TestCase):
    @mock.patch("core.storage.backends.BaseStorage.save", mock_save)
    def test_populate_data(self):
        # Test basic scenario
        generate_basic_data()



import re
import uuid
from typing import Optional


def is_valid_uuid(uuid_to_test: str, version: int = 4) -> bool:
    """
    Check if uuid_to_test is a valid UUID.

    Parameters
    ----------
    uuid_to_test : str
    version : {1, 2, 3, 4}

    Returns
    -------`
    `True` if uuid_to_test is a valid UUID, otherwise `False`.

    Examples
    --------
    >>> is_valid_uuid('c9bf9e57-1685-4c89-bafb-ff5af830be8a')
    True
    >>> is_valid_uuid('c9bf9e58')
    False
    """
    try:
        uuid_obj = uuid.UUID(uuid_to_test, version=version)
    except ValueError:
        return False

    return str(uuid_obj) == uuid_to_test


def validate_password(password: str) -> bool:
    """
    Checks if password meets the requirements and if so,
    returns a boolean. Does not sanitize password

    Passwords must be at least 12 characters and contain at least 1 from each
    of the following categories
        * Uppercase letters: A-Z.
        * Lowercase letters: a-z.
        * Numbers: 0-9.
        * Symbols: !@?&#$%

    Passwords cannot contain any whitespace characters, note that
    leading and trailing whitespace is automatically trimmed by
    the serializer.

    TODO Add features to prevent easily guessed password i.e.
        You can't use a password that:
            * Is particularly weak. Example: "password123"
            * You've used before on your account
    """

    if password is None:
        return False

    # Check if password meets length requirements
    if len(password) < 12:
        return False

    # check if password contains any form of whitespace
    whitespace_regex = r"\s"
    whitespace_match = re.search(whitespace_regex, password)

    # If whitespace is found return False
    if whitespace_match is not None:
        return False

    # Check if the password meets the character requirements using regex
    password_regex = (
        r"^(?=.*[a-z])(?=.*[A-Z])(?=.*\d)(?=.*[!@?&#$%])[A-Za-z\d!@?&#$%]{12,}$"
    )

    password_match = re.fullmatch(password_regex, password)

    # If password was matched correctly the return bool
    return password_match is not None



from cloud_scheduler.views import CronJobStatusViewset
from core.routers import CustomAPIRouter
from django.urls import include, path

app_name = "cloud_scheduler"

router = CustomAPIRouter()
router.register(prefix="", viewset=CronJobStatusViewset)

urlpatterns = [
    path("", include(router.urls)),
]



import json
import os
from typing import Dict, Optional

from django.utils import timezone
from google.cloud import tasks_v2
from google.protobuf import timestamp_pb2


def create_http_task(
    json_payload: Dict,
    scheduled_seconds_from_now: Optional[int] = None,
    task_id: Optional[str] = None,
) -> tasks_v2.Task:
    """Create an HTTP POST task with a JSON payload.
    Args:
        url: The target URL of the task.
        json_payload: The JSON payload to send.
        scheduled_seconds_from_now: Seconds from now to schedule the task for.
        task_id: ID to use for the newly created task.
        deadline_in_seconds: The deadline in seconds for task.
    Returns:
        The newly created task.
    """

    project = os.environ["GOOGLE_CLOUD_PROJECT"] or "taficasa-core-dev"
    location = "europe-west1"
    queue = "after-effects-queue"
    url = "https://taficasa-core-api-dev-scheduler-ohvkdl34jq-ew.a.run.app/"

    # Create a client.
    client = tasks_v2.CloudTasksClient()

    # Construct the task.
    task = tasks_v2.Task(
        http_request=tasks_v2.HttpRequest(
            http_method=tasks_v2.HttpMethod.POST,
            url=url,
            headers={"Content-type": "application/json"},
            body=json.dumps(json_payload).encode(),
        ),
        name=(
            client.task_path(project, location, queue, task_id)
            if task_id is not None
            else None
        ),
    )

    # Convert "seconds from now" to an absolute Protobuf Timestamp
    if scheduled_seconds_from_now is not None:
        timestamp = timestamp_pb2.Timestamp()
        timestamp.FromDatetime(
            timezone.now() + timezone.timedelta(seconds=scheduled_seconds_from_now)
        )
        task.schedule_time = timestamp

    # Use the client to send a CreateTaskRequest.
    return client.create_task(
        tasks_v2.CreateTaskRequest(
            # The queue to add the task to
            parent=client.queue_path(project, location, queue),
            # The task itself
            task=task,
        )
    )



import os

from cryptography.fernet import Fernet


def convert_bytes_to_storage_hash(bytes: bytes) -> str:
    hash_key = os.environ["STORAGE_KEY_HASH_SECRET"]
    storage_key = Fernet(hash_key).encrypt(bytes)
    return storage_key



from django.conf import settings
from django.test.runner import DiscoverRunner
from mongoengine import connect, disconnect, get_connection
from pymongo import MongoClient


class ExtendedTestRunner(DiscoverRunner):
    def setup_databases(self, **kwargs):
        # Disconnect from the current default alias db
        disconnect()

        # Setup mongo db test db
        connect(
            db=f"test_{settings.LOCAL_MONGO_DB_NAME}",
            username=settings.LOCAL_MONGO_DB_USER,
            password=settings.LOCAL_MONGO_DB_PASSWORD,
        )

        # Continue rest of db setup
        return super().setup_databases(**kwargs)

    def teardown_databases(self, old_config, **kwargs):
        # Get connection to test db and drop db
        mongo_client = MongoClient(
            f"mongodb://{settings.LOCAL_MONGO_DB_USER}:{settings.LOCAL_MONGO_DB_PASSWORD}@localhost:27017/"
        )
        mongo_client.drop_database(f"test_{settings.LOCAL_MONGO_DB_NAME}")

        # Disconnect from the test db
        disconnect()

        # Reconnect to the current default alias db
        connect(
            db=settings.LOCAL_MONGO_DB_NAME,
            username=settings.LOCAL_MONGO_DB_USER,
            password=settings.LOCAL_MONGO_DB_PASSWORD,
        )

        return super().teardown_databases(old_config, **kwargs)



from core.data.gen.organizations import generate_organizations
from core.data.gen.product_receipts import (
    ProductReceiptGeneratorStatus,
    generate_product_receipts,
)
from core.data.raw.basic import get_basic_dataset
from django.db import transaction
from supply_chain.profiles.constants import SupplyChainOrganizationType


def generate_basic_data():
    with transaction.atomic(durable=True):
        # Get organizations data
        result = generate_organizations(dataset=get_basic_dataset(), cache_result=True)

        # Get singular lender
        lender = result[SupplyChainOrganizationType.LENDER][0]
        # Generate product receipts data
        generate_product_receipts(
            merchants=result[SupplyChainOrganizationType.MERCHANT],
            endorsees=[
                [lender, distributor]
                for distributor in result[SupplyChainOrganizationType.DISTRIBUTOR]
            ],
            number_of_product_receipts=30,
            status_mix={
                ProductReceiptGeneratorStatus.CREATED: 0.25,
                ProductReceiptGeneratorStatus.IN_MERCHANT_BRANCH_VAULT: 0.15,
                ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_CENTRAL_VAULT: 0.15,
                ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_BRANCH_VAULT: 0.15,
                ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_CENTRAL_VAULT: 0.15,
                ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_BRANCH_VAULT: 0.15,
            },
            participant_matrix_strategy="product",
        )



from pathlib import Path

from authentication.roles.models import Role
from authentication.roles.services import RoleService
from authentication.users.models.auxillary import PasswordResetRequest
from authentication.users.services import UserService
from authentication.users.services.profiles import UserProfileService
from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.files.images import ImageFile
from django.db import transaction
from organization_management.branches.services import BranchService
from organization_management.organizations.services import OrganizationService
from organization_management.teams.services import TeamService
from organization_management.workflows.constants import (
    WorkflowSchemaUseCaseType,
    WorkflowTargetObject,
)
from organization_management.workflows.services.workflow_schema_services import (
    WorkflowSchemaService,
)
from organization_management.workflows.services.workflow_services import WorkflowService
from supply_chain.product_receipts.services.pickup import ProductReceiptPickupService
from supply_chain.products.services import ProductService
from supply_chain.profiles.constants import SupplyChainOrganizationType
from supply_chain.profiles.services import SupplyChainProfileService
from utils.terminal_colors import TerminalColor


def _get_logo_image(
    image_name: str, org_type: SupplyChainOrganizationType
) -> ImageFile | None:
    # Load logo image into uploaded file
    image_path = (
        Path(settings.BASE_DIR) / f"apps/core/data/static/{org_type}/{image_name}"
    )
    if image_path.exists():
        return ImageFile(open(image_path, "rb"), name=image_name)

    return None


def _get_profile_image(image_name: str) -> ImageFile | None:
    # Load logo image into uploaded file
    image_path = Path(settings.BASE_DIR) / f"apps/core/data/static/users/{image_name}"
    if image_path.exists():
        return ImageFile(open(image_path, "rb"), name=image_name)

    return None


def generate_organizations(dataset, cache_result: bool = False):
    """
    This scenario sets up data for the merchants
    """

    # Cache result if requested for now only cache supply chain profiles
    result_cache = {}

    # Generate users, orgs etc.
    with transaction.atomic():
        for data in dataset:
            # First create the organization owner
            organization_owner = get_user_model().objects.create_user(
                firstname=data["owner"]["firstname"],
                lastname=data["owner"]["lastname"],
                password=data["owner"]["password"],
                email=data["owner"]["email"],
            )
            organization_owner.email_validated = True
            organization_owner.save()
            print(
                f"{TerminalColor.OKGREEN}Successfully created organization owner for {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Create the organization for the owner
            organization = OrganizationService().create_organization(
                name=data["organization_name"],
                creator=organization_owner,
                location_data=data["branches"][data["headquarter_branch_name"]][
                    "location"
                ],
                logo=_get_logo_image(
                    image_name=data["logo_file"], org_type=data["organization_type"]
                ),
            )
            organization.is_verified = True
            organization.save()
            print(
                f"{TerminalColor.OKGREEN}Successfully created organization {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Create the proper supply chain profile
            supply_chain_profile = SupplyChainProfileService.create_profile(
                organization=organization,
                created_by=organization_owner,
                organization_type=data["organization_type"],
                product_receipt_prefix=data["product_receipt_prefix"]
                if data["organization_type"] == SupplyChainOrganizationType.MERCHANT
                else None,
            )

            # Cache supply chain profile if necessary
            if cache_result:
                if data["organization_type"] not in result_cache:
                    result_cache[data["organization_type"]] = []
                result_cache[data["organization_type"]].append(supply_chain_profile)

            print(
                f"{TerminalColor.OKGREEN}Successfully created supply chain profile {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Create all branches
            branch_obj_map = {}
            for branch_key, branch in data["branches"].items():
                if branch["is_headquarters"]:
                    # Just save and skip headquarters branch since it has already been created
                    headquarters = BranchService.get_headquarter_branch(
                        organization=organization
                    )
                    BranchService.update_branch(
                        branch=headquarters,
                        validated_data={"name": branch["name"]},
                        updated_by=organization_owner,
                    )
                    branch_obj_map[branch_key] = headquarters
                else:
                    branch_obj = BranchService.create_branch(
                        name=branch["name"],
                        organization=organization,
                        creator=organization_owner,
                        location_data=branch["location"],
                    )
                    branch_obj_map[branch_key] = branch_obj

            print(
                f"{TerminalColor.OKGREEN}Successfully created branches for {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Create all teams
            team_obj_map = {}
            for team_key, team in data["teams"].items():
                team_obj = TeamService().create_team(
                    name=team["name"],
                    organization=organization,
                    creator=organization_owner,
                )
                team_obj_map[team_key] = team_obj

            print(
                f"{TerminalColor.OKGREEN}Successfully created teams for {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Create all roles
            role_obj_map = {}
            for role_key, role in data["roles"].items():
                role_obj = RoleService.create_role(
                    name=role["name"],
                    description=role["description"],
                    organization=organization,
                    creator=organization_owner,
                    permission_types=role["permissions"],
                )
                role_obj_map[role_key] = role_obj

            # Also add the organization's default roles to the map
            for key, name in [
                ("technical_admin", "Technical Admin"),
                ("super_user", "Super User"),
            ]:
                role_obj_map[key] = Role.objects.filter(
                    name=name, organization=organization
                ).get()

            print(
                f"{TerminalColor.OKGREEN}Successfully created roles for {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Add Members to the org if necessary
            if len(data["members"]) > 0:
                OrganizationService.add_member(
                    organization=organization,
                    emails=[member["email"] for member in data["members"]],
                    added_by=organization_owner,
                )

            print(
                f"{TerminalColor.OKGREEN}Successfully added members for {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Complete user sign up
            user_obj_map = {}
            for member in data["members"]:
                member_obj = (
                    get_user_model().objects.filter(email=member["email"]).get()
                )

                # Save obj for later use
                user_obj_map[member["email"]] = member_obj

                # Roundabout way to get password, we first reset with a known password then use
                # that to complete the signup. This is done because the password for a new member is
                # normally sent to their email directly
                UserService(user=member_obj).request_password_reset()
                reset_request = PasswordResetRequest.objects.filter(
                    user=member_obj, completed=False
                ).get()
                UserService(user=member_obj).reset_password(
                    reset_id=reset_request.reset_id, new_password=member["password"]
                )

                UserService().complete_proxy_user_signup(
                    user=member_obj,
                    firstname=member["firstname"],
                    lastname=member["lastname"],
                    old_password=member["password"],
                    new_password="Admin@Password1",
                )
                member_obj.email_validated = True
                member_obj.save()

                # Update profile photo
                UserProfileService(user=member_obj).update_profile_picture(
                    new_picture=_get_profile_image(image_name=member["profile_photo"])
                )

                # Add user to their branch
                BranchService.add_member(
                    user=member_obj,
                    branch=branch_obj_map[member["branch"]],
                    added_by=organization_owner,
                )

                # Add user to any teams they are a part of
                for team in member["teams"]:
                    TeamService.add_member(
                        user=member_obj,
                        team=team_obj_map[team],
                        added_by=organization_owner,
                    )

                # Add any necessary roles to users
                for role in member["roles"]:
                    RoleService.assign_user_to_role(
                        user=member_obj,
                        role=role_obj_map[role],
                        added_by=organization_owner,
                    )
            print(
                f"{TerminalColor.OKGREEN}Successfully updated all member attributes for {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Give the org owner all permissions via a role
            RoleService.assign_user_to_role(
                user=organization_owner,
                role=role_obj_map["org_owner"],
                added_by=organization_owner,
            )
            print(
                f"{TerminalColor.OKGREEN}Successfully created super role and assigned to owner for {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Generate any workflows for the organization
            for use_case, workflow in data["workflows"].items():
                # Replace roles in workflow steps with role objects
                for step in workflow["steps"]:
                    if "owned_by_role" in step:
                        step["owned_by_role"] = role_obj_map[step["owned_by_role"]]

                WorkflowSchemaService.create_or_update_schema_version(
                    version_id=1,
                    organization=organization,
                    steps=workflow["steps"],
                    updated_by=organization_owner,
                    use_case=use_case,
                    should_publish=True,
                    published_by=organization_owner,
                )
            print(
                f"{TerminalColor.OKGREEN}Successfully created all workflows for {data['organization_name']}{TerminalColor.ENDC}"
            )

            # Generate depots products only for merchant
            if data["organization_type"] == SupplyChainOrganizationType.MERCHANT:
                for product in data["products"]:
                    # Load product image into uploaded file
                    image_path = (
                        Path(settings.BASE_DIR)
                        / "apps/core/data/static/products"
                        / product["image"]
                    )
                    image_file = ImageFile(
                        open(image_path, "rb"), name=product["image"]
                    )
                    # Create product
                    product = ProductService.create_product(
                        merchant=supply_chain_profile,
                        name=product["name"],
                        merchant_product_id=product["merchant_product_id"],
                        created_by=organization_owner,
                        image=image_file,
                        unit=product["unit"],
                    )
                    # Autocomplete product workflow
                    WorkflowService.auto_complete_workflow(
                        workflow=product.publish_approval_workflow,
                        target_object=WorkflowTargetObject.NONE,
                        target_object_id=str(product.product_id),
                        use_case=WorkflowSchemaUseCaseType.PRODUCT_PUBLISH,
                    )
                    # Publish it
                    ProductService.publish_product(
                        product=product, published_by=organization_owner
                    )

                for branch in branch_obj_map.values():
                    ProductReceiptPickupService.create_pickup_depot(
                        supply_chain_profile=supply_chain_profile,
                        name=branch.name,
                        merchant_depot_id=f"depot-{branch.name}",
                        branch=branch,
                        created_by=organization_owner,
                    )
            # Generate access requests for the bank only
            if data["organization_type"] == SupplyChainOrganizationType.LENDER:
                for request in data["access_requests"]:
                    RoleService.start_role_request_workflow(
                        role=role_obj_map[request["role"]],
                        users=[user_obj_map[user] for user in request["users"]],
                        organization=organization,
                        requested_by=organization_owner,
                    )
            print(
                f"{TerminalColor.OKGREEN}Successfully created products for {data['organization_name']}{TerminalColor.ENDC}"
            )

    return result_cache



from rest_framework.views import exception_handler as default_exception_handler


def api_exception_handler(exception, context):
    # Call REST framework's default exception handler first,
    # to get the standard error response.
    response = default_exception_handler(exception, context)

    # Now add the HTTP status code to the response.
    if response is not None:
        response.data["status_code"] = response.status_code

    # TODO If exception should be logged then do so
    if getattr(exception, "should_log", None):
        # Do some logging
        pass

    # TODO If exception should raise alert
    if getattr(exception, "raise_alert", None):
        # Do some alerting
        pass

    return response



from django.urls import reverse
from rest_framework.test import APIClient


class CustomAPIClient(APIClient):
    def __init__(self, login_data):
        super().__init__()
        self._login_data = login_data
        self._login_data["remember"] = False

    def update_login_data(self, field: str, value: str):
        self._login_data[field] = value

    def custom_login(self):
        login_response = self.post(
            reverse("authentication:login"), self._login_data, format="json"
        )
        return login_response

    def custom_logout(self):
        return self.post(reverse("authentication:logout"))



from enum import StrEnum

from core.mixins import ModelChoiceMixin
from django.db import models


class CustomCurrencyField(models.DecimalField):
    """
    Defines a custom subclass of the decimal field
    to be uniformly used for currecny/money throughout app
    """

    def __init__(self, *args, **kwargs):
        kwargs["decimal_places"] = 2
        kwargs["max_digits"] = 18
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["decimal_places"]
        del kwargs["max_digits"]

        return name, path, args, kwargs


class CustomRateField(models.DecimalField):
    """
    Defines a custom subclass of the decimal field
    to be uniformly used for rates (e.g interest rate)
    throughout app
    """

    def __init__(self, *args, **kwargs):
        kwargs["decimal_places"] = 3
        kwargs["max_digits"] = 6
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["decimal_places"]
        del kwargs["max_digits"]

        return name, path, args, kwargs


class SupportedCurrencies(ModelChoiceMixin, StrEnum):
    """
    Represents supported currencies
    """

    NIGERIAN_NAIRA = "ngn"
    US_DOLLAR = "usd"

    @classmethod
    def max_name_length(cls) -> int:
        return 3


class SupportedFloatingInterestRates(ModelChoiceMixin, StrEnum):
    """
    Represents supported widely accepted floating interest rates
    """

    NIGERIAN_PRIME_LENDING_RATE = "nigerian_prime_lending_rate"

    @classmethod
    def max_name_length(cls) -> int:
        return 128



from typing import Optional

from rest_framework import exceptions, status


class BaseAPIErrorCode:
    code: str
    detail: str


class APIServiceError(exceptions.APIException):
    def __init__(
        self,
        error_code: BaseAPIErrorCode,
        status_code: int = status.HTTP_400_BAD_REQUEST,
    ):
        """
        General error class for service classes, uses status code 400

        params - message: message to show the api user
                 describing the error

                 code: unique str error code for this error
        """
        super().__init__(
            detail=error_code.detail,
            code=error_code,
        )
        self.status_code = status_code
        self.error_code = error_code


class APIValidationError(exceptions.APIException):
    status_code = status.HTTP_400_BAD_REQUEST

    def __init__(
        self,
        message: str,
    ):
        """
        General error class for validation errors, uses status code 400

        params - message: message to show the api user
                 describing the error
        """
        super().__init__(
            detail=message,
            code="api_validation_error",
        )


class GenericInternalError(exceptions.APIException):
    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR

    def __init__(self, message: str, code: str = "generic_internal_error"):
        """
        Generic error class for internal server errors, uses status code 500

        params - message: message to show the api user
                 describing the error

                 code: unique str error code for this error
        """

        super().__init__(
            detail=message,
            code=code,
        )


class InvalidRequestError(exceptions.APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = "Invalid Request."
    default_code = "invalid_request"

    def __init__(self, message: Optional[str] = None, code: Optional[int] = None):
        """
        General error class for cases when a request
        contains invalid data or cannot be processed.
        """

        super().__init__(
            detail=message or self.default_detail,
            code=code or self.default_code,
        )



import itertools
import random
from decimal import Decimal
from enum import IntEnum
from typing import Dict, List, Literal, Optional

import numpy as np
import supply_chain.product_receipts.exceptions as product_receipt_exceptions
from authentication.users.models import User
from core.currency import SupportedCurrencies
from core.exceptions import APIServiceError
from django.db import transaction
from django.utils import timezone
from organization_management.branches.services import BranchService
from organization_management.organizations.services import OrganizationService
from organization_management.workflows.constants import (
    WorkflowSchemaUseCaseType,
    WorkflowTargetObject,
)
from organization_management.workflows.services import WorkflowService
from supply_chain.product_receipts.models import (
    ProductReceipt,
    ProductReceiptAssignment,
)
from supply_chain.product_receipts.services import ProductReceiptService
from supply_chain.products.services import ProductService
from supply_chain.profiles.constants import SupplyChainOrganizationType
from supply_chain.profiles.models import SupplyChainProfile
from utils.terminal_colors import TerminalColor


class ProductReceiptGeneratorStatus(IntEnum):
    CREATED = 0
    IN_MERCHANT_BRANCH_VAULT = 5
    IN_FIRST_ENDORSEMENT_CENTRAL_VAULT = 10
    IN_FIRST_ENDORSEMENT_BRANCH_VAULT = 15
    IN_SECOND_ENDORSEMENT_CENTRAL_VAULT = 20
    IN_SECOND_ENDORSEMENT_BRANCH_VAULT = 25


def _apply_endorsement_status(
    product_receipts: List[ProductReceipt], should_back_date: bool = False
) -> None:
    for idx, product_receipt in enumerate(product_receipts):
        try:
            # Get current endorsee who owns product receipt
            product_receipt_owner = ProductReceiptService.get_owner(
                product_receipt=product_receipt
            )

            # Create endorsement workflow
            ProductReceiptService.start_endorsement_workflow(
                product_receipt=product_receipt,
                supply_chain_profile=product_receipt_owner,
                created_by=product_receipt_owner.organization.owner,
            )

            # Autocomplete workflow
            assignment = ProductReceiptAssignment.objects.filter(
                product_receipt=product_receipt,
                supply_chain_profile=product_receipt_owner,
            ).get()
            WorkflowService.auto_complete_workflow(
                workflow=assignment.endorsement_workflow,
                target_object=WorkflowTargetObject.PRODUCT_RECEIPT,
                target_object_id=str(product_receipt.product_receipt_id),
                use_case=WorkflowSchemaUseCaseType.PRODUCT_RECEIPT_ENDORSEMENT,
            )
            # Endorse product receipt
            ProductReceiptService.endorse_product_receipt(
                product_receipt=product_receipt,
                supply_chain_profile=product_receipt_owner,
                endorsed_by=product_receipt_owner.organization.owner,
                back_date_endorsement=product_receipt.created_on
                + timezone.timedelta(days=random.choice([1, 1, 1, 1, 2, 3]))
                if should_back_date
                else None,
            )
        except APIServiceError as e:
            if e.error_code == product_receipt_exceptions.ProductReceiptError0014:
                # This error is bad since it means the auto approval failed
                raise e
            else:
                # For any other errors re raise
                raise e


def _apply_branch_vault_status(product_receipts: List[ProductReceipt]) -> None:
    for product_receipt in product_receipts:
        try:
            # Get current endorsee who owns product receipt
            product_receipt_owner = ProductReceiptService.get_owner(
                product_receipt=product_receipt
            )
            #  Get all branches used for picking random branch to assign
            branches = OrganizationService(
                organization=product_receipt_owner.organization
            ).get_branches()

            # Create assignment workflow
            ProductReceiptService.start_branch_assignment_workflow(
                product_receipt=product_receipt,
                supply_chain_profile=product_receipt_owner,
                created_by=product_receipt_owner.organization.owner,
            )
            # Autocomplete workflow
            assignment = ProductReceiptAssignment.objects.filter(
                product_receipt=product_receipt,
                supply_chain_profile=product_receipt_owner,
            ).get()
            WorkflowService.auto_complete_workflow(
                workflow=assignment.branch_assignment_workflow,
                target_object=WorkflowTargetObject.PRODUCT_RECEIPT,
                target_object_id=str(product_receipt.product_receipt_id),
                use_case=WorkflowSchemaUseCaseType.PRODUCT_RECEIPT_ASSIGNMENT_TO_BRANCH,
            )
            # Complete assignment
            ProductReceiptService.assign_product_receipt_to_branch(
                product_receipt=product_receipt,
                supply_chain_profile=product_receipt_owner,
                assigned_by=product_receipt_owner.organization.owner,
                branch=random.choice(branches),
            )
        except APIServiceError as e:
            if e.error_code == product_receipt_exceptions.ProductReceiptError0026:
                # This error is bad since it means the auto approval failed
                raise e
            else:
                # For any other errors re raise
                raise e


def _create_product_receipts(
    merchant: SupplyChainProfile,
    endorsees: List[SupplyChainProfile],
    created_by: User,
    qty: int,
    back_date_list: Optional[List[timezone.datetime]] = None,
) -> list[ProductReceipt]:
    # Set up price details and precalculate random prices
    prices = [50, 100, 200, 500, 1000, 10000, 25000, 50000, 100000, 500000]
    weights = [0.15, 0.2, 0.2, 0.15, 0.1, 0.05, 0.05, 0.05, 0.025, 0.025]
    random_prices = random.choices(population=prices, weights=weights, k=qty)

    # Set up quantity details and precalculate random quantities
    quantities = [50000, 1000, 200, 500, 100000, 10000, 25000, 500000, 100, 50]
    random_quantities = random.choices(population=quantities, weights=weights, k=qty)

    product_receipts = []
    # Get merchants current products and distribute evenly amongst product receipts
    products = ProductService.get_products(
        merchant=merchant, include_unpublished="False"
    )
    for idx in range(qty):
        product_receipts.append(
            ProductReceiptService.create_product_receipt(
                merchant=merchant,
                holders=[endorsee.profile_id for endorsee in endorsees],
                product=products[idx % len(products)],
                quoted_price=Decimal(random_prices[idx]),
                quoted_price_currency=SupportedCurrencies.NIGERIAN_NAIRA,
                quoted_price_expiry=timezone.now() + timezone.timedelta(days=180),
                created_by=created_by,
                total_quantity=random_quantities[idx],
                documents=[],
                back_dated_creation=back_date_list[idx % len(back_date_list)]
                if back_date_list
                else None,
            )
        )
    return product_receipts


def generate_product_receipts(
    merchants: List[SupplyChainProfile],
    endorsees: List[List[SupplyChainProfile]],
    number_of_product_receipts: int = 1,
    status_mix: Dict[ProductReceiptGeneratorStatus, int] = {
        ProductReceiptGeneratorStatus.CREATED: 1
    },
    participant_matrix_strategy: Literal["product", "linear"] = "product",
    backdate_start: Optional[timezone.datetime] = None,
):
    """Generates product receipts based on passed params

    merchants -- The list of merchants to generate product receipts for
    bulk_agents -- The list of bulk agents to generate product receipts for
    customers -- The list of customers(distributor/end user) to create product receipts for
    number_of_product_receipts -- The number of product receipts to generate per merchant/customer pair
    status_mix -- The ratio for applying different statuses to the generated statuses
    participant_matrix_strategy -- The strategy for generating merchant/customer pairs
    backdate_start -- If this is set then the product receipts created will be backdated in an equal amount
                     starting at this date with a daily frequency
    """

    # Create participant matrix
    if participant_matrix_strategy == "linear":
        participant_matrix = zip(merchants, endorsees)
    else:
        # default to cross product always
        participant_matrix = itertools.product(merchants, endorsees)

    # if any error occurs during this process the user should be able to fix and retry
    with transaction.atomic():
        # Create a list of random dates starting from the backdate start date to the present
        back_date_list = None
        if backdate_start:
            days_delta = (timezone.now() - backdate_start).days
            back_date_list = [
                backdate_start + timezone.timedelta(days=idx)
                for idx in range(days_delta)
            ]

        back_date_start = 0
        back_date_stop = min(number_of_product_receipts, len(back_date_list))
        product_receipts = []
        for pair in participant_matrix:
            # Create product receipts first
            product_receipts.extend(
                _create_product_receipts(
                    merchant=pair[0],
                    endorsees=pair[1],
                    created_by=pair[0].organization.owner,
                    qty=number_of_product_receipts,
                    back_date_list=back_date_list[back_date_start:back_date_stop],
                )
            )

            # Calculate next start and stop indices for back date list
            if back_date_list:
                # If there is still more dates in the list then use them
                if len(back_date_list) > back_date_stop:
                    back_date_start = back_date_stop
                    back_date_stop = min(
                        back_date_start + number_of_product_receipts,
                        len(back_date_list),
                    )
                # If not then reset
                else:
                    back_date_start = 0
                    back_date_stop = min(
                        number_of_product_receipts, len(back_date_list)
                    )

        print(f"{TerminalColor.OKGREEN}Successfully created all product receipts")

        # Shuffle product receipts before partitioning
        random.shuffle(product_receipts)

        # Calculate Normalized chunks
        status_keys = list(status_mix.keys())

        # Get split indices from status mix percentages
        split_indices_raw = [
            int(len(product_receipts) * status_mix[status]) for status in status_keys
        ][:-1]
        split_indices = np.cumsum(split_indices_raw)

        # Split product receipts and store in distribution dict
        product_receipts_split = np.split(product_receipts, split_indices)
        distribution = {
            status_keys[idx]: list(product_receipts_split[idx])
            for idx in range(len(status_keys))
        }

        # Apply status
        for status in distribution:
            if status == ProductReceiptGeneratorStatus.IN_MERCHANT_BRANCH_VAULT:
                _apply_branch_vault_status(
                    product_receipts=distribution[
                        ProductReceiptGeneratorStatus.IN_MERCHANT_BRANCH_VAULT
                    ]
                    + distribution[
                        ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_CENTRAL_VAULT
                    ]
                    + distribution[
                        ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_BRANCH_VAULT
                    ]
                    + distribution[
                        ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_CENTRAL_VAULT
                    ]
                    + distribution[
                        ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_BRANCH_VAULT
                    ]
                )
                print(
                    f"{TerminalColor.OKGREEN}Successfully applied merchant branch vault status"
                )
            elif (
                status
                == ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_CENTRAL_VAULT
            ):
                _apply_endorsement_status(
                    product_receipts=(
                        distribution[
                            ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_CENTRAL_VAULT
                        ]
                        + distribution[
                            ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_BRANCH_VAULT
                        ]
                        + distribution[
                            ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_CENTRAL_VAULT
                        ]
                        + distribution[
                            ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_BRANCH_VAULT
                        ]
                    ),
                    should_back_date=backdate_start != None,
                )
                print(
                    f"{TerminalColor.OKGREEN}Successfully applied first endorsement central vault status"
                )
            elif (
                status
                == ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_BRANCH_VAULT
            ):
                _apply_branch_vault_status(
                    product_receipts=(
                        distribution[
                            ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_BRANCH_VAULT
                        ]
                        + distribution[
                            ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_CENTRAL_VAULT
                        ]
                        + distribution[
                            ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_BRANCH_VAULT
                        ]
                    )
                )
                print(
                    f"{TerminalColor.OKGREEN}Successfully applied first endorsement branch vault status"
                )
            elif (
                status
                == ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_CENTRAL_VAULT
            ):
                _apply_endorsement_status(
                    product_receipts=(
                        distribution[
                            ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_CENTRAL_VAULT
                        ]
                        + distribution[
                            ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_BRANCH_VAULT
                        ]
                    )
                )
                print(
                    f"{TerminalColor.OKGREEN}Successfully applied second endorsement central vault status"
                )
            elif (
                status
                == ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_BRANCH_VAULT
            ):
                _apply_endorsement_status(
                    product_receipts=distribution[
                        ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_BRANCH_VAULT
                    ]
                )
                print(
                    f"{TerminalColor.OKGREEN}Successfully applied second endorsement branch vault status"
                )


def product_receipts_data(number: int):
    """
    This scenario sets up a fairly small/basic amount of
    data for product receipts and skips all the prep for products and merchant registrations

    For now it assumes that the basic user/organization scenario has been
    ran
    """

    # Get the relevant supply chain profiles for each users org
    merchant_profiles = list(
        SupplyChainProfile.objects.filter(
            organization_type=SupplyChainOrganizationType.MERCHANT
        ).all()
    )
    lender_profiles = list(
        SupplyChainProfile.objects.filter(
            organization_type=SupplyChainOrganizationType.LENDER
        ).all()
    )
    distributor_profiles = list(
        SupplyChainProfile.objects.filter(
            organization_type=SupplyChainOrganizationType.DISTRIBUTOR
        ).all()
    )

    # Scramble the returned lender profiles so that every time this is ran we get a slightly
    # different mix
    random.shuffle(lender_profiles)

    # Create endorsees list
    if len(distributor_profiles) >= len(lender_profiles):
        endorsees = [
            [lender_profiles[idx % len(lender_profiles)], distributor]
            for idx, distributor in enumerate(distributor_profiles)
        ]
    else:
        endorsees = [
            [
                lender,
                distributor_profiles[idx % len(distributor_profiles)],
            ]
            for idx, lender in enumerate(lender_profiles)
        ]

    # Generate product receipts
    generate_product_receipts(
        merchants=merchant_profiles,
        endorsees=endorsees,
        number_of_product_receipts=number,
        status_mix={
            ProductReceiptGeneratorStatus.CREATED: 0.25,
            ProductReceiptGeneratorStatus.IN_MERCHANT_BRANCH_VAULT: 0.15,
            ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_CENTRAL_VAULT: 0.15,
            ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_BRANCH_VAULT: 0.15,
            ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_CENTRAL_VAULT: 0.15,
            ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_BRANCH_VAULT: 0.15,
        },
        participant_matrix_strategy="product",
        backdate_start=timezone.now() - timezone.timedelta(days=365),
    )



from enum import StrEnum

from core.mixins import ModelChoiceMixin


class SupportedCountries(ModelChoiceMixin, StrEnum):
    """
    Represents supported countries
    """

    NIGERIA = "nga"

    def max_name_length(cls) -> int:
        return 16



import functools
from typing import Callable, Dict, List, ParamSpec, TypeVar

from django.conf import settings

T = TypeVar("T")
P = ParamSpec("P")


def check_effect_arguments(
    args: List[str], kwargs: Dict[str, str | int | None]
) -> bool:
    allowed_types = [int, str, None]
    # Check args
    for arg in args:
        if type(arg) not in allowed_types:
            return False

    # Check kwargs
    for key, value in kwargs.items():
        if type(key) != str or type(value) not in allowed_types:
            return False

    return True


class AfterEffectRegistry:
    def __init__(self):
        self._registered_effects = {}
        self._queued_tasks = []

    @property
    def queued_tasks(self):
        return self._queued_tasks

    @property
    def registered_effects(self):
        return self._registered_effects

    def __call__(self, app_name):
        def decorator(func: Callable[P, T]) -> Callable[P, T]:
            effect_unique_name = f"{app_name}:{func.__name__}"
            self._registered_effects[effect_unique_name] = func

            @functools.wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                # Check types before creating task
                if not check_effect_arguments(args=args, kwargs=kwargs):
                    raise ValueError(
                        "After Effect Functions can only take arguments that accept str, int or None values"
                    )

                if settings.LOCAL or settings.TESTING:
                    # Call task immediately
                    func(*args, **kwargs)
                else:
                    # Delegate to gcp task queue which will be processed in
                    # order after response is received from view
                    self._queued_tasks.append(
                        {
                            "json_payload": {
                                "effect": f"{app_name}:{func.__name__}",
                                "effect_args": args,
                                "effect_kwargs": kwargs,
                            },
                            "scheduled_seconds_from_now": 60,
                        }
                    )

            return wrapper

        return decorator

    def __get__(self, obj, objtype=None):
        return self._registered


after_effect = AfterEffectRegistry()



from typing import List, Tuple


class ModelChoiceMixin:
    """
    This mixin is used to extend python enums so
    they can be used in model choice fields.
    """

    @classmethod
    def as_choices(cls) -> List[Tuple[int, str]]:
        return list(map(lambda c: (c.value, c.name), cls))

    def as_choice(self) -> Tuple[int, str]:
        return (self.value, self.name)

    @classmethod
    def max_name_length(cls) -> int:
        raise NotImplementedError()



from rest_framework import serializers


class AfterEffectSerializer(serializers.Serializer):
    effect = serializers.CharField()
    effect_args = serializers.ListField(child=serializers.CharField())
    effect_kwargs = serializers.DictField()



import secrets


def generate_temporary_password():
    """
    Returns a temporary password of size 12 chars
    """
    symbols = "!@?&#$%"
    lower_letters = "abcdefghjkmnpqrstuvwxyz"
    upper_letters = "ABCDEFGHJKLMNPQRSTUVWXYZ"
    digits = "23456789"

    max_uppers = 3
    max_lowers = 3
    max_digits = 4
    max_symbols = 2

    temp = ""

    lowers_added = 0
    uppers_added = 0
    digits_added = 0
    symbols_added = 0

    counter = 0

    while counter < 12:
        # Evaluate which chars to choose from
        char_choices = ""

        if uppers_added < max_uppers:
            char_choices += upper_letters
        if lowers_added < max_lowers:
            char_choices += lower_letters
        if digits_added < max_digits:
            char_choices += digits
        if symbols_added < max_symbols:
            char_choices += symbols

        random_char = secrets.choice(char_choices)

        if random_char.isalpha():
            if random_char.isupper():
                if uppers_added < max_uppers:
                    uppers_added += 1
                    temp += random_char
                    counter += 1
            else:
                if lowers_added < max_lowers:
                    lowers_added += 1
                    temp += random_char
                    counter += 1
        else:
            if random_char.isdigit():
                if digits_added < max_digits:
                    digits_added += 1
                    temp += random_char
                    counter += 1
            else:
                if symbols_added < max_symbols:
                    symbols_added += 1
                    temp += random_char
                    counter += 1

    return temp


def generate_temp_id(length=16):
    """
    Returns a random id of specified length.
    id consists of alphanumeric chars only
    """
    charset = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789"

    random_id = "".join([secrets.choice(charset) for _ in range(length)])

    return random_id



import random
from pathlib import Path
from typing import Iterator, List

import organization_management.workflows.exceptions as workflow_exceptions
import supply_chain.products.exceptions as product_exceptions
from core.exceptions import APIServiceError
from django.conf import settings
from django.core.files.images import ImageFile
from django.db import transaction
from organization_management.workflows.constants import (
    WorkflowSchemaUseCaseType,
    WorkflowStepType,
    WorkflowTargetObject,
)
from organization_management.workflows.services import (
    WorkflowSchemaService,
    WorkflowService,
)
from supply_chain.products.models import Product
from supply_chain.products.services import ProductService
from supply_chain.profiles.models import SupplyChainProfile

PRODUCTS = [
    {
        "name": "Cement",
        "merchant_product_id": "0001",
        "image": "cement.jpeg",
        "unit": "bags",
    },
    {
        "name": "Cooking Oil",
        "merchant_product_id": "0002",
        "image": "cooking_oil.jpeg",
        "unit": "liters",
    },
    {
        "name": "Sugar",
        "merchant_product_id": "0003",
        "image": "sugar.jpg",
        "unit": "tons",
    },
    {
        "name": "Petrol",
        "merchant_product_id": "0004",
        "image": "petrol.jpg",
        "unit": "liters",
    },
    {
        "name": "Cassava",
        "merchant_product_id": "0005",
        "image": "cassava.jpg",
        "unit": "tons",
    },
]


def _get_products_data(qty: int) -> Iterator[Product]:
    # For each call shuffle products data
    random.shuffle(PRODUCTS)

    for idx in range(min(qty, len(PRODUCTS))):
        yield PRODUCTS[idx]


def generate_products(
    merchants: List[SupplyChainProfile], products_per_merchant: int = 1
) -> List[Product]:
    # if any error occurs during this process the user should be able to fix and retry
    with transaction.atomic():
        # Create Product and publishing schema for all merchants
        created_products = []
        for merchant in merchants:
            try:
                WorkflowSchemaService.create_or_update_schema_version(
                    version_id=1,
                    organization=merchant.organization,
                    steps=[
                        {
                            "step_title": "Owner Approval Step",
                            "step_description": "Generated step for owner approval",
                            "step_type": WorkflowStepType.APPROVAL,
                            "step_order": 1,
                        }
                    ],
                    updated_by=merchant.organization.owner,
                    use_case=WorkflowSchemaUseCaseType.PRODUCT_PUBLISH,
                    should_publish=True,
                    published_by=merchant.organization.owner,
                )
            except APIServiceError as e:
                if (
                    e.error_code == workflow_exceptions.WorkflowError0001
                    or workflow_exceptions.WorkflowError0002
                ):
                    # This is fine since it means that the schema already exists
                    pass
                else:
                    raise e

            for product_data in _get_products_data(qty=products_per_merchant):
                try:
                    # Load product image into uploaded file
                    image_path = (
                        Path(settings.BASE_DIR)
                        / "apps/core/data/static"
                        / product_data["image"]
                    )
                    image_file = ImageFile(
                        open(image_path, "rb"), name=product_data["image"]
                    )
                    # Create product
                    product = ProductService.create_product(
                        merchant=merchant,
                        name=product_data["name"],
                        merchant_product_id=product_data["merchant_product_id"],
                        created_by=merchant.organization.owner,
                        image=image_file,
                        unit=product_data["unit"],
                    )
                    # Autocomplete product workflow
                    WorkflowService.auto_complete_workflow(
                        workflow=product.publish_approval_workflow,
                        target_object=WorkflowTargetObject.NONE,
                        target_object_id=str(product.product_id),
                        use_case=WorkflowSchemaUseCaseType.PRODUCT_PUBLISH,
                    )
                    # Publish it
                    ProductService.publish_product(
                        product=product, published_by=merchant.organization.owner
                    )
                    created_products.append(product)
                except APIServiceError as e:
                    if e.error_code == product_exceptions.ProductError0001:
                        # This error is fine since it just means the product already exists
                        pass
                    elif e.error_code == product_exceptions.ProductError0005:
                        # This error is bad since it means the auto approval failed
                        raise e
                    else:
                        # For any other errors re raise
                        raise e
        return created_products



from core.data.gen.organizations import generate_organizations
from core.data.gen.product_receipts import (
    ProductReceiptGeneratorStatus,
    generate_product_receipts,
)
from core.data.raw.distributors import get_distributors_dataset, get_flux_distributor
from core.data.raw.lenders import get_flux_bank, get_lenders_dataset
from core.data.raw.merchants import get_flux_merchant, get_merchants_dataset
from django.db import transaction
from django.utils import timezone
from supply_chain.profiles.constants import SupplyChainOrganizationType


def generate_demo_data_no_analytics():
    with transaction.atomic(durable=True):
        # Get organizations data
        dataset = [get_flux_merchant(), get_flux_distributor(), get_flux_bank()]
        generate_organizations(dataset=dataset)


def generate_demo_data_with_analytics():
    with transaction.atomic(durable=True):
        # Get organizations data
        dataset = (
            get_merchants_dataset() + get_distributors_dataset() + get_lenders_dataset()
        )
        result = generate_organizations(dataset=dataset, cache_result=True)

        # Get singular lender
        lender = result[SupplyChainOrganizationType.LENDER][0]
        # Generate product receipts data
        generate_product_receipts(
            merchants=result[SupplyChainOrganizationType.MERCHANT],
            endorsees=[
                [lender, distributor]
                for distributor in result[SupplyChainOrganizationType.DISTRIBUTOR]
            ],
            number_of_product_receipts=10,
            status_mix={
                ProductReceiptGeneratorStatus.CREATED: 0,
                ProductReceiptGeneratorStatus.IN_MERCHANT_BRANCH_VAULT: 0,
                ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_CENTRAL_VAULT: 0.60,
                ProductReceiptGeneratorStatus.IN_FIRST_ENDORSEMENT_BRANCH_VAULT: 0.40,
                ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_CENTRAL_VAULT: 0,
                ProductReceiptGeneratorStatus.IN_SECOND_ENDORSEMENT_BRANCH_VAULT: 0,
            },
            participant_matrix_strategy="product",
            backdate_start=timezone.now() - timezone.timedelta(days=365),
        )



from collections import OrderedDict

from cloud_scheduler.services import get_job_status
from core.routers import CustomAPIRoute
from rest_framework import mixins
from rest_framework import status as http_status
from rest_framework import viewsets
from rest_framework.response import Response


class CronJobStatusViewset(
    mixins.CreateModelMixin,
    viewsets.GenericViewSet,
):
    custom_api_routes = [
        CustomAPIRoute(
            url="job/status",
            url_args=OrderedDict(),
            url_name="job-status",
            methods=["create"],
        ),
    ]

    def create(self, request, *args, **kwargs):
        """
        Description: Used for getting the status
                     of a google cloud scheduler job

        URL Args: - None

        URL Kwargs: - None

        Request Data: None

        Returns: TODO
        """

        status = get_job_status(name=request.data["name"])

        return Response(
            data={"status": status},
            status=http_status.HTTP_200_OK,
        )



from django.apps import apps
from django.contrib.auth import get_user_model
from rest_framework import serializers, validators


class UserEmailExistsSerializerValidator:
    def __call__(self, value):
        queryset = get_user_model().objects.filter(email=value)

        if not validators.qs_exists(queryset):
            raise serializers.ValidationError(
                "This field must be an email of a valid user."
            )


class EqualFieldsValidator:
    """
    This validator checks if all the fields passed in are equal

    Should be applied to the serializer class and not an individual
    field
    """

    def __init__(self, fields):
        self.fields = fields

    def __call__(self, attrs):
        hashset = set()
        for field in self.fields:
            hashset.add(attrs[field])

        if len(hashset) != 1:
            raise serializers.ValidationError(
                f"The fields - {','.join(self.fields)} must all have the same value"
            )



from core.data.gen.basic import generate_basic_data
from core.data.gen.demo import (
    generate_demo_data_no_analytics,
    generate_demo_data_with_analytics,
)
from core.data.gen.product_receipts import product_receipts_data



from functools import wraps

from django.contrib.auth import get_user_model
from django.core.exceptions import PermissionDenied


def ensure_same_user(view_func):
    """
    A decorator that ensures that the user passed
    in the request using the key 'user' matches
    the user that is actually authenticated and that is
    initiating the request.

    Raises Permission Denied Error
    """

    @wraps(view_func)
    def wrapper(request, *args, **kwargs):
        # try to get the user passed in the request data
        queryset = get_user_model().filter(email=request.data["user"])

        if not bool(queryset):
            raise PermissionDenied

        user_instance = queryset.get()

        if str(user_instance.id) != str(request.user.id):
            raise PermissionDenied

        return view_func(request, *args, **kwargs)



from django.conf import settings
from django.utils.translation import gettext_lazy as _
from knox.auth import TokenAuthentication
from rest_framework import exceptions, request


class CookieBasedTokenAuthentication(TokenAuthentication):
    def _get_auth_cookie_token(self, request):
        if settings.AUTH_COOKIE_NAME in request.COOKIES:
            # Note str needs to be encoded to work with knox auth method
            auth_token = request.COOKIES[settings.AUTH_COOKIE_NAME].encode("utf-8")
            return auth_token
        msg = _("Authentication Cookie Not Provided")
        raise exceptions.AuthenticationFailed(msg)

    def authenticate(self, request: request.Request):
        # Use regular token based auth for local env excluding testing and for
        # calls made to internal urls
        if (settings.LOCAL and not settings.TESTING) or request.path.startswith(
            "/v1/internal"
        ):
            response = super().authenticate(request)
            # Special check to make sure that on non local env only
            # bot users can successfully log in without cookies
            if response and not settings.LOCAL and not response[0].is_bot:
                raise exceptions.AuthenticationFailed(_("Only Bot users can login"))
            return response
        else:
            auth_cookie_token = self._get_auth_cookie_token(request=request)
            user, auth_token = self.authenticate_credentials(auth_cookie_token)
            if user.is_bot:
                raise exceptions.AuthenticationFailed(
                    _("Bot users can't login with cookies")
                )
            return user, auth_token



from enum import StrEnum


class URLArgumentName(StrEnum):
    USER = "userid"
    ORGANIZATION = "organization_id"
    ORGANIZATION_PUBLIC = "public_id"
    TEAM = "team_id"
    ROLE = "role_id"
    ROLE_REQUEST = "role_request_id"
    BRANCH = "branch_id"
    WORKFLOW = "workflow_id"
    WORKFLOW_SCHEMA = "workflow_schema_id"
    WORKFLOW_SCHEMA_USE_CASE = "use_case"
    PRODUCT = "product_id"
    PRODUCT_CATEGORY = "product_category_id"
    NOTIFICATION = "notification_id"
    PRODUCT_RECEIPT = "product_receipt_id"
    PICKUP_REQUEST = "pickup_request_id"
    MERCHANT_REGISTRATION = "merchant_registration_id"
    SUPPLY_CHAIN_PROFILE = "profile_id"
    LOAN_APPLICATION = "application_id"
    VERIFICATION = "verification_id"
    STEP_ORDER = "step_order"



"""
A module where constants used throughout the 
project should be defined
"""


MAX_EMAIL_LENGTH = 256
MAX_NAME_FIELD_LENGTH = 256
MAX_ORG_ENTITY_NAME_LENGTH = 512

MAX_UUID_LENGTH = 30

UNAMBIGUOUS_CHARS = [
    "B",
    "C",
    "D",
    "F",
    "G",
    "H",
    "J",
    "K",
    "M",
    "P",
    "R",
    "T",
    "V",
    "W",
    "X",
    "Y",
    "2",
    "3",
    "4",
    "6",
    "7",
    "8",
    "9",
]



from typing import Tuple, final

import names
from authentication.users.models import User
from core.testing.constants import DEFAULT_TEST_PASSWORD
from django.contrib.auth import get_user_model
from organization_management.organizations.models import Organization

from .supply_chain import SupplyChainDataMixin


class BaseTestDataClient:
    """
    This client should provide an easy way of generating test data
    across the app
    """

    def __init__(self) -> None:
        self.created_users = set()
        self.created_products = set()

    def get_test_organization_profile(self) -> Tuple[User, Organization]:
        # Import services in functions to avoid circular imports
        from organization_management.organizations.services import OrganizationService
        from supply_chain.profiles.services import SupplyChainProfileService

        # Create a new user
        random_name = names.get_full_name()
        while random_name in self.created_users:
            random_name = names.get_full_name()

        firstname, lastname = random_name.split(" ")
        user_data = {
            "email": f"{firstname}{lastname}@test.com",
            "firstname": firstname,
            "lastname": lastname,
            "password": DEFAULT_TEST_PASSWORD,
        }

        # Create User to be used in tests
        user: User = get_user_model().objects.create_user(**user_data)
        user.email_validated = True
        user.save()

        organization = OrganizationService().create_organization(
            "Test Organization",
            user,
            location_data={
                "street_address": "No 1 lokogoma street",
                "country": "NG",
                "locality": "Wuse",
                "administrative_area": "Abuja",
            },
        )

        organization.is_verified = True
        organization.save()

        # Update the created users set
        self.created_users.add(random_name)

        return user, organization


@final
class TestDataClient(SupplyChainDataMixin, BaseTestDataClient):
    """
    This is the final data client that should include all mixins
    """



"""
This module defines job states and Job status
it provides a function to convert numeric state/status codes to their corresponding names
"""


class JobStatus:
    OK = 0
    CANCELLED = 1
    UNKNOWN = 2
    INVALID_ARGUMENT = 3
    DEADLINE_EXCEEDED = 4
    NOT_FOUND = 5
    ALREADY_EXISTS = 6
    PERMISSION_DENIED = 7
    RESOURCE_EXHAUSTED = 8
    FAILED_PRECONDITION = 9
    ABORTED = 10
    OUT_OF_RANGE = 11
    UNIMPLEMENTED = 12
    INTERNAL = 13
    UNAVAILABLE = 14
    DATA_LOSS = 15
    UNAUTHENTICATED = 16


def convert_status_code(status_code):
    status_mapping = {
        JobStatus.OK: "OK",
        JobStatus.CANCELLED: "CANCELLED",
        JobStatus.UNKNOWN: "UNKNOWN",
        JobStatus.INVALID_ARGUMENT: "INVALID_ARGUMENT",
        JobStatus.DEADLINE_EXCEEDED: "DEADLINE_EXCEEDED",
        JobStatus.NOT_FOUND: "NOT_FOUND",
        JobStatus.ALREADY_EXISTS: "ALREADY_EXISTS",
        JobStatus.PERMISSION_DENIED: "PERMISSION_DENIED",
        JobStatus.RESOURCE_EXHAUSTED: "RESOURCE_EXHAUSTED",
        JobStatus.FAILED_PRECONDITION: "FAILED_PRECONDITION",
        JobStatus.ABORTED: "ABORTED",
        JobStatus.OUT_OF_RANGE: "OUT_OF_RANGE",
        JobStatus.UNIMPLEMENTED: "UNIMPLEMENTED",
        JobStatus.INTERNAL: "INTERNAL",
        JobStatus.UNAVAILABLE: "UNAVAILABLE",
        JobStatus.DATA_LOSS: "DATA_LOSS",
        JobStatus.UNAUTHENTICATED: "UNAUTHENTICATED",
    }
    return status_mapping.get(status_code, "UNKNOWN")


class JobState:
    STATE_UNSPECIFIED = 0
    ENABLED = 1
    PAUSED = 2
    DISABLED = 3
    UPDATE_FAILED = 4


def convert_job_state(state):
    state_mapping = {
        JobState.STATE_UNSPECIFIED: "STATE_UNSPECIFIED",
        JobState.ENABLED: "ENABLED",
        JobState.PAUSED: "PAUSED",
        JobState.DISABLED: "DISABLED",
        JobState.UPDATE_FAILED: "UPDATE_FAILED",
    }
    return state_mapping.get(state, "STATE_UNSPECIFIED")



from dataclasses import dataclass
from typing import Dict, List, Optional, OrderedDict, Tuple

from django.core.exceptions import ImproperlyConfigured
from django.urls import path, re_path
from rest_framework.routers import BaseRouter
from rest_framework.viewsets import GenericViewSet


@dataclass(frozen=True)
class CustomAPIRoute:
    # URL path string with zero-indexed format strings
    # e.g. {0} which will be swapped out for url_args
    url: str
    # Mapping of url arg name to optional django path converter
    # if path converter is not specified, the arg is defaulted to
    # the '[^/.]+' regex capturing group.
    url_args: OrderedDict[str, Optional[str]]
    # Unique name of URL. This is what you can pass into reverse
    # to get back the final URL string, like so
    # reverse("<app_label>:<url_name>")
    url_name: Dict[str, str]
    # List of methods that were implemented on the viewset
    methods: List[str]


class CustomAPIRouter(BaseRouter):
    # Default implementation, TODO check if I need to change
    def get_default_basename(self, viewset):
        return ""

    def _get_routes(self, viewset) -> List[CustomAPIRoute]:
        if not hasattr(viewset, "custom_api_routes"):
            raise ImproperlyConfigured(
                f"To use CustomAPIRouter, Viewset - {viewset.__name__} "
                "must set the custom_api_routes attribute"
            )

        # Validate that urls are set correctly in custom api routes
        # and do not have starting forward slashes
        for route in viewset.custom_api_routes:
            if route.url.startswith("/"):
                raise ImproperlyConfigured(
                    f"URL - {route.url} on CustomAPIRoute for Viewset - {viewset.__name__} "
                    "cannot start with a forward slash (/)"
                )
        return viewset.custom_api_routes

    def _get_route_action(self, action):
        """
        Returns a map between a http mthod and the
        corresponding viewset action(method)
        """
        if action == "list" or action == "retrieve":
            return {"get": action}
        elif action == "create" or action == "update":
            return {"post": action}
        elif action == "partial_update":
            return {"patch": action}
        elif action == "destroy":
            return {"delete": action}

        raise ImproperlyConfigured(
            f"Viewset method - {action} set in custom_api_routes not recognized!"
        )

    def _get_route_mapping(self, methods):
        mapping = dict()
        for method in methods:
            mapping.update(self._get_route_action(method))

        return mapping

    def _get_lookup_url(self, url: str, url_args: Dict[str, str]) -> Tuple[str, bool]:
        lookup_url = url
        used_regex = False
        used_converter = False

        for idx, (arg, converter) in enumerate(url_args.items()):
            if converter:
                if used_regex:
                    raise ImproperlyConfigured(
                        f"For URL {url}, You must use either path convereters or "
                        "the default regex exclusively"
                    )
                replacement = f"(<{converter}:{arg}>)"
                used_converter = True
            else:
                if used_converter:
                    raise ImproperlyConfigured(
                        f"For URL - {url}, You must use either path convereters or "
                        "the default regex exclusively"
                    )
                replacement = f"(?P<{arg}>[^/.]+)"
                used_regex = True

            replacement_string = f"{{{idx}}}"
            lookup_url = lookup_url.replace(replacement_string, replacement)

        return lookup_url, used_regex

    def get_urls(self):
        """
        Use the registered viewsets to generate a list of URL patterns.
        """
        url_patterns = []
        for prefix, viewset, basename in self.registry:
            routes = self._get_routes(viewset)
            for route in routes:
                # Build the url pattern
                lookup_url, used_regex = self._get_lookup_url(route.url, route.url_args)
                if lookup_url:
                    final_url = "{prefix}/{lookup}/".format(
                        prefix=prefix,
                        lookup=lookup_url,
                    )
                else:
                    final_url = "{prefix}/".format(
                        prefix=prefix,
                    )

                # In the case no prefix is defined we need to make sure we don't have
                # a leading url
                final_url = final_url.lstrip("/")

                initkwargs = {
                    "basename": route.url_name,
                }
                view = viewset.as_view(
                    self._get_route_mapping(route.methods), **initkwargs
                )
                if used_regex:
                    regex_path = r"^{0}$".format(final_url)
                    url_patterns.append(re_path(regex_path, view, name=route.url_name))
                else:
                    url_patterns.append(path(final_url, view, name=route.url_name))

        return url_patterns

    def register_in_bulk(self, prefix: str, viewsets: List[GenericViewSet]):
        """
        Convenience method to register a lot of viewsets at once
        to one common prefix
        """

        for viewset in viewsets:
            self.register(prefix, viewset)



from after_effects.registry import after_effect
from after_effects.tasks import create_http_task


class AfterEffectsMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.
        while len(after_effect.queued_tasks):
            task = after_effect.queued_tasks.pop(0)
            create_http_task(
                json_payload=task["json_payload"],
                scheduled_seconds_from_now=task["scheduled_seconds_from_now"],
            )
        return response



from typing import Optional

import google.auth
from cloud_scheduler.status import convert_job_state, convert_status_code
from google.cloud import scheduler_v1


def create_scheduler_job(
    url: str,
    method: str,
    job_name: str,
    schedule: str,
    token: str,
    description: Optional[str] = None,
    data: Optional[bytes] = None,
) -> scheduler_v1.Job:
    """
    This function creates a cloud scheduler job on google cloud.

    Args:
    - url (str): The full URI path that the request will be sent to. This string must begin with either "http://" or "https://".
    - method (str): The HTTP method to be used for the request ('GET', 'POST', 'PUT', 'DELETE', etc.).
    - job_name (str): The name of the Cloud Scheduler job.
    - schedule (str): The cron schedule for the job in the form 'MINUTE HOUR DAY_OF_MONTH MONTH DAY_OF_WEEK' e.g (* * * * *).
    - token (str): The Authorization token for the requested endpoint
    - description (Optional[str]): Description of the job (default is None).
    - data (Optional[bytes]]): The data to be sent in the request body (default is None).

    Returns:
    - response (dict): The response from creating the Cloud Scheduler job.

    Documentation:
    - https://cloud.google.com/python/docs/reference/cloudscheduler/latest/google.cloud.scheduler_v1.services.cloud_scheduler.CloudSchedulerClient

    """
    credentials, project_id = google.auth.default()

    # Create a client
    client = scheduler_v1.CloudSchedulerClient(credentials=credentials)

    location = "europe-west1"
    parent = f"projects/{project_id}/locations/{location}"

    # Get HttpMethod enum value
    http_method = getattr(scheduler_v1.HttpMethod, method.upper())

    http_target = {
        "uri": url,
        "http_method": http_method,
        "headers": {"Authorization": f"Token {token}"},
    }

    # Add data to http_target body if request method is post or put
    if (
        http_method in (scheduler_v1.HttpMethod.POST, scheduler_v1.HttpMethod.PUT)
        and data
    ):
        http_target["body"] = data

    job = {
        "name": f"{parent}/jobs/{job_name}",
        "description": description,
        "http_target": http_target,
        "schedule": schedule,
        "time_zone": "UTC",
    }

    response = client.create_job(parent=parent, job=job)

    return response


def get_job_status(name: str):
    """
    This function returns the status of a google cloud scheduler job
    """

    credentials, _ = google.auth.default()

    # Create a client
    client = scheduler_v1.CloudSchedulerClient(credentials=credentials)

    # Initialize request
    request = scheduler_v1.GetJobRequest(name=name)

    # Make the request
    response = client.get_job(request=request)

    # return the job state is status is not present
    if not response.status.code:
        return convert_job_state(response.state)

    return convert_status_code(response.status.code)



from after_effects.views import AfterEffectRunViewset
from core.routers import CustomAPIRouter
from django.urls import include, path

app_name = "after_effects"

router = CustomAPIRouter()
router.register_in_bulk(
    prefix="after_effects",
    viewsets=[AfterEffectRunViewset],
)


urlpatterns = [
    path("", include(router.urls)),
]



from setuptools import setup, find_packages

setup(
    name="taficasa_shared_apps",
    version="0.1.0", 
    author="Excel Okechukwu",
    author_email="excel@taficasa.com",
    description="This package contains folders and apps needed within TafiCasa apps",
    long_description=open('README.rst').read(),
    long_description_content_type="text/x-rst",
    license="Custom License",  
    classifiers=[
        "Framework :: Django",
        "Programming Language :: Python :: 3",
    ],
    keywords="django utilities cloud scheduler shared components",
    packages=find_packages(), 
    python_requires=">=3.8",
    install_requires=[
        "Django==4.2.2",
        'django-rest-knox==4.2.0',
        'djangorestframework==3.14.0',
        'google-cloud-tasks==2.16.3',
        'google-cloud-scheduler==2.13.3',
    ],
   
)



from enum import StrEnum


class SendgridVerifiedSender(StrEnum):
    """
    Represents verified senders in sendgrid
    """

    SUPPORT = "support@taficasa.com"


class SengridEmailTemplate(StrEnum):
    """
    Email templates registered in sendgrid
    """

    PROXY_WELCOME = "d-0f7604d9f9714613b0d935470ff9f5d9"
    PASSWORD_RESET = "d-2310b065d2004e6ebfdd65a69de7fdef"
    PASSWORD_CHANGED = "d-c8cfe8afc49942b0885f08dfc312b151"
    EMAIL_VERIFICATION = "d-18d95840cfac477a93de3bac9b8975a6"
    WELCOME_EMAIL = "d-8ce326146bb64ffd910f34e51a74c91c"



from authentication.permissions.constants import (
    PERMISSION_EXPLANATION_MAP,
    PermissionType,
)
from django.core import checks


@checks.register()
def run_access_control_permissions_check(app_configs, **kwargs):
    errors = []
    for permission in PermissionType:
        if (
            permission not in PERMISSION_EXPLANATION_MAP
            and permission != PermissionType.UNKNOWN
        ):
            errors.append(
                checks.Error(
                    f"{permission} missing in Permission Type Explanations",
                    hint=(
                        "Please make sure every permission type defined in PermissionType "
                        "has an explanation in PERMISSION_EXPLANATION_MAP The only exception "
                        "is the UNKOWN Permission Type."
                    ),
                )
            )
            break

    return errors



import io
from typing import Any, Optional

import magic
from django.core.files import File
from django.core.files.base import ContentFile
from PIL import Image

SUPPORTED_MIME_TO_FORMAT = {
    "image/jpeg": "JPEG",
    "image/png": "PNG",
}


def resize_image(
    uploaded_image: File,
    max_height: Optional[int] = None,
    max_width: Optional[int] = None,
) -> Optional[ContentFile]:
    uploaded_image_bytes = io.BytesIO(uploaded_image.read())
    # Check image format
    mime_type = magic.from_buffer(uploaded_image_bytes.read(2048), mime=True)
    if mime_type not in SUPPORTED_MIME_TO_FORMAT:
        return None
    uploaded_image_format = SUPPORTED_MIME_TO_FORMAT[mime_type]

    # Read in image
    image_obj = Image.open(uploaded_image_bytes)
    width, height = image_obj.size

    # Get ratio for resizing
    if not max_height:
        resize_ratio = max_width / width
    elif not max_width:
        resize_ratio = max_height / height
    else:
        resize_ratio = min((max_width / width), (max_height / height))

    # Calc new sizes and round to nearest int
    new_width = round(resize_ratio * width)
    new_height = round(resize_ratio * height)

    # Resize image
    resized_image = image_obj.resize(
        size=(new_width, new_height), resample=Image.Resampling.LANCZOS
    )

    # Return a file that can be saved directly to db
    final_image_bytes = io.BytesIO()
    resized_image.save(final_image_bytes, format=uploaded_image_format, quality=95)
    final_image = ContentFile(final_image_bytes.getvalue(), name=uploaded_image.name)

    return final_image
